/**
 * SLPv2 DA Database Management
 *
 * We implemented the database function within Java2. We didn't use
 * a standard DBMS for simplicity and efficiency. The whole database
 * is organized as red-black tree (provided by Java)
 *
 *  (1) the key is ltag+URL, assume each service has a unique URL, 
 *      and the same service can be registered using different 
 *      languaage (ltag)
 *  (2) the value at each node is an Entry class, which keeps all the
 *      information about the service.
 *
 * (c) Columbia University, 2001-2004, All Rights Reserved.
 * Author: Weibin Zhao
 */

import java.io.*;
import java.util.*;

public class Database {

    da daf;
    TreeMap table;
    slpMsgComposer composer;
    Vector matchedEntry;
    ByteArrayOutputStream b;
    DataOutputStream d;
    int totalMatch;

    Database(da daf) {
        this.daf = daf;
        table = new TreeMap();
        composer = new slpMsgComposer();
        matchedEntry = new Vector(10);
        b = new ByteArrayOutputStream();
        d = new DataOutputStream(b);
    }

    // return the number of entries in the database
    public synchronized int size() {
        return table.size();
    }

    public synchronized TreeMap table() {
        return table;
    }

    //------------------------------------------------
    // save database to either the stdout or the file
    //------------------------------------------------
    public synchronized void saveDatabase(BufferedWriter o) {
        Iterator values = table.values().iterator();
        while (values.hasNext()) {
            Entry e = (Entry) values.next();
            e.prtEntry(daf, o);
        }
    }

    //------------------------------------------
    // load data form file to internal database
    //------------------------------------------
    public synchronized void loadDatabase(String dbase) {
        String line, ltag, type, url, scope, acceptDA, attr = "";
        int lifetime;
        long versionTS, arrivalTS, acceptTS;
        boolean deleted = false;
        try {
            BufferedReader in = 
                new BufferedReader(new FileReader(dbase));
            while ((line = in.readLine()) != null) {
                StringTokenizer st = new StringTokenizer(line);
                deleted = Boolean.valueOf(st.nextToken()).booleanValue();
                ltag = st.nextToken();
                type = st.nextToken();
                url  = st.nextToken();
                lifetime = Integer.parseInt(st.nextToken());
                scope = st.nextToken();
                versionTS = Long.parseLong(st.nextToken());
                arrivalTS = Long.parseLong(st.nextToken());
                acceptDA = st.nextToken();
                acceptTS = Long.parseLong(st.nextToken());
                if (st.hasMoreTokens()) attr = st.nextToken();
                addEntry(deleted, ltag, type, url, lifetime, scope, attr, 
                         Const.fresh_flag, versionTS, arrivalTS, 
                         acceptDA, acceptTS);
            }
            in.close();
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    //-------------------------------------------
    // check lifetime and remove expired entries
    //-------------------------------------------
    public synchronized void rmExpiredEntry() {
        long currtime = System.currentTimeMillis();
        Iterator keys = table.keySet().iterator();
        while (keys.hasNext()) {
            String k = (String) keys.next();
            Entry  e = (Entry)  table.get(k);
            if (currtime > (e.getArrivalTS() + e.getLifetime()*1000)) {
                keys.remove();
            }
        }
    }

    //---------------------------------------------------------
    // for "SrvReg"
    // add a new entry of service registration to the database
    // or replace/update its previous registration
    //---------------------------------------------------------
    public synchronized int addEntry(boolean deleted, String ltag, String type,
        String url, int lifetime, String scope, String attr, int reg_flag,
        long versionTS, long arrivalTS, String acceptDA, long acceptTS) {
        if (!table.containsKey(ltag+url)) {
            if ((reg_flag & Const.fresh_flag) == 0) { // incermental SrvReg
                return Const.INVALID_UPDATE;
            }
            table.put(ltag+url, new Entry());
        }
        Entry e = (Entry) table.get(ltag+url);
        return e.update(deleted, ltag, url, type, lifetime, scope, attr, 
                        reg_flag, versionTS, arrivalTS, acceptDA, acceptTS);
    }

    //---------------------------------------------------
    // for "SrvDeReg"
    // remove the entry with the key: ltag+url (tag=="")
    // or delete some attributes of this entry (tag!="")
    //---------------------------------------------------
    public synchronized int rmEntry(String ltag, String url, String scope, 
                String tag, long versionTS, String acceptDA, long acceptTS) {
        if (table.containsKey(ltag+url)) {
            Entry e = (Entry) table.get(ltag+url);
            if (!scope.equalsIgnoreCase(e.getScope())) {
                return Const.SCOPE_NOT_SUPPORTED;
            }
            e.deletion(tag, versionTS, acceptDA, acceptTS);
        }
        return Const.OK;
    }

    //----------------------------------------------------------------------
    // for "SrvTypeRqst"
    // get the list of service types for specified scope & naming authority
    //----------------------------------------------------------------------
    public synchronized String getServiceTypeList(String na, String scope) {
        Vector typelist = new Vector(5);
        Iterator values = table.values().iterator();
        while (values.hasNext()) {
            Entry e = (Entry) values.next();
            if (!e.getDeleted() &&                       // nor deleted
                scope.equalsIgnoreCase(e.getScope()) &&  // match scope 
                (na.equals("*") ||                       // NA wildcard 
                 na.equalsIgnoreCase(e.getNA())) &&      // match NA
                !typelist.contains(e.getType())) {
                typelist.addElement(e.getType());
            }
        }
        StringBuffer tl = new StringBuffer();
        for (int i=0; i<typelist.size(); i++) {
            String s = (String)typelist.elementAt(i);
            if (tl.length() > 0) tl.append(",");
            tl.append(s);
        }
        return tl.toString();
    }

    //-----------------------------------------------------------
    // for "SrvRqst"
    // find the matched URLs with (type, scope, predicate, ltag)
    // return: error code (short)
    //         number of matched URLs (short)
    //         URL blocks (decided by previous #URL)
    //-----------------------------------------------------------

    public int getTotalMatch() {
        return totalMatch;
    }

    public Vector getMatchedEntry() {
        return matchedEntry;
    }

    public synchronized byte[] getMatchedURL(String type, String scope,
                String pred, String ltag, Vector ssExtList) {
        byte[] buf = null;
        int ecode = Const.OK;
        if (!Util.shareString(daf.getScope(), scope, ",")) {
            ecode = Const.SCOPE_NOT_SUPPORTED;
        } 
        // obtain matched entries
        matchedEntry.clear();
        Iterator values = table.values().iterator();
        while (values.hasNext()) {
            Entry e = (Entry) values.next();
            if (e.match(type, scope, pred, ltag)) matchedEntry.addElement(e);
        }
        totalMatch = matchedEntry.size();
        // filter matched entries
        for (int i=0; i<ssExtList.size(); i++) {
            SelectSortExt ss = (SelectSortExt) ssExtList.elementAt(i);
            if (ss.getID() == Const.SelectExt) {
                int bound = ss.getBound();
                if (bound < matchedEntry.size()) matchedEntry.setSize(bound);
            } else if (ss.getID() == Const.SortExt) {
                sortFilter(ss.getKey());
            }
        }
        // write matched URLs to buffer
        b.reset();
        try {
            d.writeShort(ecode);                // error code 
            d.writeShort(matchedEntry.size());  // URL count
            if (ecode == Const.OK) {            // fill in matched URLs
                for (int i=0; i<matchedEntry.size(); i++) {
                    Entry e = (Entry) matchedEntry.elementAt(i);
                    d.writeByte(0);
                    d.writeShort(e.getLifetime());
                    d.writeShort(e.getURL().length());
                    d.writeBytes(e.getURL());
                    d.writeByte(0);
                }
            }
            buf = b.toByteArray();
        } catch (Exception ex) {
            System.err.println("getMatchedURL:"+ex);
        }
        return buf;
    }

    private void sortFilter(String key) {
        int size = matchedEntry.size();
        if (size <= 1) return; // no need to sort
        StringTokenizer st = new StringTokenizer(key, ",");
        int nkey = st.countTokens();
        String[]  keys   = new String[nkey];
        int[]     types  = new int[nkey];
        int[]     orders = new int[nkey];
        Float[]   vals   = new Float[nkey];
        for (int i=0; i<nkey; i++) {
            String unit = st.nextToken();
            StringTokenizer st1 = new StringTokenizer(unit, ":");
            if (st1.countTokens() < 3) {
                System.out.println("Incorrect sort key list");
            }
            keys[i] = st1.nextToken().trim();  // key
            String s = st1.nextToken().trim(); // type
            if (s.equalsIgnoreCase("s")) {
                types[i] = Const.StringSort;
            } else {
                types[i] = Const.NumberSort;
            }
            s = st1.nextToken().trim();        // order
            if (s.equals("+")) {
                orders[i] = Const.IncreasingOrder;
            } else {
                orders[i] = Const.DecreasingOrder;
            }
            if (st1.hasMoreTokens()) {         // reference value
                vals[i] = new Float(st1.nextToken().trim());
            } else {
                vals[i] = null;
            }
        }
        Vector se = new Vector(size);
        for (int i=0; i<size; i++) {
            Entry e = (Entry)matchedEntry.elementAt(i);
            se.addElement(new SortEntry(e, keys, types, orders, vals));
        }
        Collections.sort(se);
        matchedEntry.clear();
        for (int i=0; i<size; i++) {
            SortEntry s = (SortEntry) se.elementAt(i);
            matchedEntry.addElement(s.getEntry());
        }
    }

    //------------------------------------------------------
    // for "AttrRqst"
    // Return: attrbute list for the URL/"service type"
    //         format: (attr1=value1),(attr2=value2)
    // if tag != "", return ONLY those attributes in tag
    //------------------------------------------------------
    public synchronized String getAttrList(String url, String scope,
                                String tag, String ltag) {
        if (table.containsKey(ltag+url)) {
            Entry e = (Entry) table.get(ltag+url);
            if (!e.getDeleted() &&
                scope.equalsIgnoreCase(e.getScope())) {
                return e.getAttr(tag);
            } else {
                return "";
            }
        } else {        // Not URL, try servive type
            return typeAttrList(url, scope, tag, ltag);
        }
    }

    private synchronized String typeAttrList(String type, String scope, 
                                String tag, String ltag) {
        StringBuffer attrList = new StringBuffer();
        Iterator values = table.values().iterator();
        while (values.hasNext()) {
            Entry e = (Entry) values.next();
            if (!e.getDeleted() &&
                type.equalsIgnoreCase(e.getType()) && 
                scope.equalsIgnoreCase(e.getScope()) &&
                ltag.equalsIgnoreCase(e.getLtag())) {
                String s = e.getAttr(tag);
                if (attrList.length() > 0) attrList.append(",");
                attrList.append(s);
            }
        }
        return attrList.toString();
    }

    //------------------------------------------------------------------
    // find new states based on selective/complete & (rdaList, rtsList)
    // sort new states on their accept IDs and return in TreeMap
    //------------------------------------------------------------------
    public synchronized Vector findNewStates(String rscope,
                 Vector rdaList, Vector rtsList, int etrpType) {

        Vector tmp = new Vector();
        Iterator values = table.values().iterator();
        while (values.hasNext()) { 
            Entry  e     = (Entry) values.next();
            String ada   = e.getAcceptDA();
            long   ats   = e.getAcceptTS();
            int    index = rdaList.indexOf(ada); // a requested subset?
            long   rts   = 0;
            if (index != -1) { // it is a requested subset, find rts
                rts = ((Long) rtsList.get(index)).longValue();
            }
            if (Util.shareString(rscope, e.getScope(), ",")) { 
                if (index != -1 && ats > rts ||
                    index == -1 && etrpType == Const.complete) { 
                    tmp.addElement(e);
                }
            }
        }
        Collections.sort(tmp);
        return tmp;
    }
}
