/**
 * SLPv2 DA (directory agent) with mesh-enhanced capability
 * It can be invoked as follows:
 * (1) java da 
 *     # use default scope, database and summary file
 *     # DAAdvert multicast is on (default)
 *     # DAAdvert has "mesh-enhanced" keyword (default)
 *     # use selective anti-entropy (default)
 * (2) java -Dscope=x da  
 *     # set DA scopes, scopes should be seperated by comma, no space
 * (3) java -Ddbase=y da  
 *     # set database file name
 * (4) java -Dsummary=z da  
 *     # set summary file name
 * (5) java -Dmcast=no da 
 *     # turn off DAAdvert multicast
 *     # simulate the condition where multicast is NOT supported
 * (6) java -Dmesh=no da 
 *     # turn off "mesh-enhanced" keyword in DAAdvert
 *     # function as a non-mesh-emhanced DA
 * (7) java -Dmode=complete da 
 *     # use complete anti-entropy (default is selective)
 * (8) java -Dload=yes da 
 *     # load data from file (default is not loading data from file)
 * (9) java -Dsave=yes da 
 *     # save data to file (default is not saving data to file)
 *
 * (c) Columbia University, 2001-2004, All Rights Reserved.
 * Author: Weibin Zhao
 */

import java.util.*;
import java.io.*;
import java.net.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;

public class da extends JFrame implements ActionListener {

    static int width = 600;
    static int height = 350;

    public static void main(String args[]) {
        try {
            boolean mcast = true;
            boolean mesh_enhanced = true;
            boolean loaddata = false;
            boolean savedata = false;
            int     etrpType;

            // we need a trick to get the FQDN of the host
            String ipaddr = (InetAddress.getLocalHost()).getHostAddress();
            String hname  = (InetAddress.getByName(ipaddr)).getHostName();

            // set URL for DAAdvert 
            String ip = System.getProperty("ip");
            if (ip == null) ip = ipaddr;

            // set scope for this DA
            String scope = System.getProperty("scope");
            if (scope == null) scope = Const.defaultScope;

            // database file
            String dbase = System.getProperty("dbase");
            if (dbase == null) dbase = Const.defaultDbase;

            // summary file
            String summary = System.getProperty("summary");
            if (summary == null) summary = Const.defaultSummary;

            // multicast DAAdvert or not?
            String s = System.getProperty("mcast");
            if (s != null && s.equalsIgnoreCase("no")) mcast = false;

            // carry "mesh-enhnaced" attribute keyword in DAAdvert or not?
            s = System.getProperty("mesh");
            if (s != null && s.equalsIgnoreCase("no")) mesh_enhanced = false;

            // use selective or complete anti-entropy
            s = System.getProperty("mode");
            if (s != null && s.equalsIgnoreCase("complete")) {
                etrpType = Const.complete;
            } else {
                etrpType = Const.selective;
            }

            // load data or not?
            s = System.getProperty("load");
            if (s != null && s.equalsIgnoreCase("yes")) loaddata = true;

            // save data or not?
            s = System.getProperty("save");
            if (s != null && s.equalsIgnoreCase("yes")) savedata = true;

            da f = new da(hname, ip, scope, dbase, summary, mcast, 
                          mesh_enhanced, etrpType, loaddata, savedata);
            f.setSize(width, height);
            f.setForeground(Color.black);
            f.setBackground(Color.lightGray);
            f.show();
        } catch (Exception e) {
            System.err.println("main:"+e);
            System.exit(1);
        }
    }

    JMenuBar   mbar;
    JMenu      fileMenu, funcMenu, setupMenu, helpMenu;
    JMenuItem  exitItem, aboutItem, dataItem, summaryItem, saveItem, loadItem, 
               confItem, peerItem, advertItem, daItem;
    JLabel     urlLabel;
    JTextField urlField;
    JTextArea  textarea;
    TreeMap  peerTable;    // key: absolute domain name, value: Peer class
    TreeMap  summaryTable; // summary of database & DAAdvert 
    TreeMap  rqstLog;      // anti-entropy request log (waiting for SrvAck)
    Database database;
    String   myfqdn, myurl;
    String   myscope, myattr="", myspi="", mydbase, mysummary;
    int      myts;
    long     startTime; // starting time of anti-entropy
    boolean  mcast, mesh_enhanced, busy;
    slpMsgComposer  composer;
    slpUdpHandler   udp;
    java.util.Timer mainTimer;
    TimerTask       mainTimerTask;
    int             daAdvertTimer, dbCheckTimer, keepaliveTimer;
    int             etrpType;   // anti-entropy type: selective or complete
    int             xid = 0;    // used in AntiEtrpRqst
    boolean  savedata;
    
    da(String hname, String ip, String scope, String dbase, String summary,
                boolean mcast, boolean mesh_enhanced, int etrpType, 
                boolean loaddata, boolean savedata) {
        super("Welcome to mSLP DA <"+hname+">");
        createGUI();

        // DA parameters
        this.mcast = mcast;
        this.mesh_enhanced = mesh_enhanced;
        this.etrpType = etrpType;
        this.savedata = savedata;
        if (mesh_enhanced) myattr = "mesh-enhanced";
        myfqdn = hname;
        myurl  = Const.DAAdvert_Rqst+"://"+ip;
        myscope = scope;
        myts = (int) (System.currentTimeMillis()/1000);
        mydbase   = dbase;
        mysummary = summary;
        busy = false;
        
        // data structure
        database = new Database(this);
        composer = new slpMsgComposer();
        peerTable   = new TreeMap();
        summaryTable = new TreeMap();
        rqstLog = new TreeMap();
        if (loaddata) loadData();  // load database, summary and advert
        long sts = 0;
        Summary s = (Summary) summaryTable.get(myfqdn);
        if (s != null)  sts = s.getAcceptTS();
        summaryTable.put(myfqdn, 
                 new Summary(myfqdn, myurl, myscope, myts, myattr, sts));

        // TCP and UDP server
        (new daTcpServer(this)).start();
        daUdpServer t = new daUdpServer(this);
        udp = t.getUdp();
        t.start();

        // timer initilization
        daAdvertTimer  = Const.mainTimer_interval; // advertise first time
        dbCheckTimer   = Const.dbCheck_interval   + Const.mainTimer_interval;
        keepaliveTimer = Const.keepalive_interval + Const.mainTimer_interval;
        mainTimer = new java.util.Timer();
        mainTimerTask  = new TimerTask() {
            public void run() { scheduleTask(); }
        };
        mainTimer.scheduleAtFixedRate(mainTimerTask, 
                Const.mainTimer_delay, Const.mainTimer_interval);

        if (mcast) multicastDAAdvert();
    }

    private synchronized void scheduleTask() {
        // Timer for multicast DAAdvert (say once a minute)
        if (mcast) { // only if multicast is supported
            daAdvertTimer -= Const.mainTimer_interval;
            if (daAdvertTimer <= 0) {
                multicastDAAdvert();
                daAdvertTimer = Const.daAdvert_interval;
            }
        }

        // Timer for purging expired data in database (say once a day)
        dbCheckTimer -= Const.mainTimer_interval;
        if (dbCheckTimer <= 0) {
            database.rmExpiredEntry();
            dbCheckTimer = Const.dbCheck_interval;
        }

        // Remove expired peers in peerTable (say once several minutes)
        rmExpiredPeer();

        // Keepalive timer for avtive peers (say once a minute)
        keepaliveTimer -= Const.mainTimer_interval;
        if (keepaliveTimer <= 0) {
            sendKeepAlive();
            keepaliveTimer = Const.keepalive_interval;
        }
    }

    private void multicastDAAdvert() {
        byte[] buf = composer.DAAdvert(0, Const.mcast_flag, "en", 
                        myts, myurl, myscope, myattr, myspi);
        String ret = udp.send(buf, Const.mcast_addr, Const.port);
        if (ret != null) append("multicast DAAdvert: "+ret);
    }

    private synchronized void rmExpiredPeer() {
        Iterator keys = peerTable.keySet().iterator();
        while (keys.hasNext()) {
            String k = (String) keys.next();
            Peer   p = (Peer) peerTable.get(k);
            if (!myfqdn.equalsIgnoreCase(p.getFQDN()) && 
                p.getAliveTimer() < 0) {
                System.out.println(">>> Remove " + p.getFQDN() + 
                                   ",  keepalive timeout");
                p.closeTcp();
                keys.remove();
            }
        }
    }

    private synchronized void sendKeepAlive() {
        byte[] buf = composer.DAAdvert(0, Const.normal_flag, "en", 
                        myts, myurl, myscope, myattr, myspi);
        Iterator values = peerTable.values().iterator();
        while (values.hasNext()) {
            Peer p = (Peer) values.next();
            if (!myfqdn.equalsIgnoreCase(p.getFQDN())) {  // this DA itself
                (p.getTcp()).send(buf, buf.length);
            }
        }
    }

    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == exitItem) {
            if (savedata) saveData();  // save data before exit
            System.exit(0);
        } else if (e.getSource() == dataItem) {
            drawLine();
            String s1 = "Deleted LangTag SrvType URL LifeTime Scope ";
            String s2 = "versionTS arrivalTS acceptDA acceptTS Attribute";
            append(s1+s2);
            database.saveDatabase(null);
        } else if (e.getSource() == summaryItem) {
            displaySummary();
        } else if (e.getSource() == advertItem) {
            displayAdvert();
        } else if (e.getSource() == saveItem) {
            saveData();
            append("data saved");
        } else if (e.getSource() == loadItem) {
            if (loadData()) append("data loaded");
        } else if (e.getSource() == confItem) {
            displayMyState();
        } else if (e.getSource() == peerItem) {
            displayPeerState();
        } else if (e.getSource() == daItem) {
            unicastDAAdvert(urlField.getText());
        }
    }    

    public boolean getMeshFlag() {
        return mesh_enhanced;
    }

    public String getScope() {
        return myscope;
    }

    public String getURL() {
        return myurl;
    }

    public String getFQDN() {
        return myfqdn;
    }

    public String getAttr() {
        return myattr;
    }

    public String getSPI() {
        return myspi;
    }

    public int getTS() {
        return myts;
    }

    public synchronized TreeMap getPeerTable() {
        return peerTable;
    }

    public synchronized Database getDatabase() {
        return database;
    }

    public void append(String s) {
        textarea.append(s + "\n");
        int line = textarea.getLineCount();
        try {
            int pos = textarea.getLineStartOffset(line-1);
            textarea.setCaretPosition(pos);
        } catch (Exception ex) {
            System.out.println("appen:" + ex);
        }
    }

    //-----------------------------------------------------------------------
    // unicast a DAAdvert to the destDA, simulate multicast is not supported
    //-----------------------------------------------------------------------
    private void unicastDAAdvert(String destDA) {
        if (destDA.equals("")) {
            append("please give To-URL for dest DA");
            return;
        }
        byte[] buf = composer.DAAdvert(0, Const.normal_flag, "en", 
                        myts, myurl, myscope, myattr, myspi);
        String ret = udp.send(buf, destDA, Const.port);
        if (ret == null) {
            append("unicast DAAdvert --> " + destDA);
        } else {
            append(ret);
        }
    }

    //------------------------------------------------------------
    // (1) create a peering connection to "peer" if needed
    // (2) add the "peer" to "peerTable", and create summary entry
    // (3) send DAAdvert(s) & AntiEtrpRqst to "peer" if it is meshDA
    //-------------------------------------------------------------
    public synchronized void DAAdvertHandler(String peer, String peer_url, 
                String scope, String attr, int bts, slpTcpHandler tcp, 
                byte[] buf, boolean origDAAdvert) {
        if (!Util.shareString(myscope, scope, ",")) return; // not a peer
        if (attr.lastIndexOf("mesh-enhanced") < 0) return; // non-mesh DA 
        if (peerTable.size() == 0) {
            startTime = System.currentTimeMillis();
        }
        Peer p = (Peer) peerTable.get(peer);
        if (p != null) {                               // an existing peer
            if (origDAAdvert) KeepAlive_Handler(peer); // keepalive
            return;
        }
        // now it is a new peer: p == null and share scopes
        Summary s = (Summary) summaryTable.get(peer);
        if (s == null) {
            summaryTable.put(peer, 
                    new Summary(peer, peer_url, scope, bts, attr, 0));
        } else {
            if (bts > s.getBootTS()) summaryTable.put(peer, 
               new Summary(peer, peer_url, scope, bts, attr, s.getAcceptTS()));
        }
        if (tcp == null) { // need to create a peering connection 
            s = (Summary) summaryTable.get(peer);
            if (s.getPflag()) return;
            s.setPflag();
            tcp = createPeerConn(peer);
            if (tcp == null) {
                append("Cannot create peering connection to " + peer);
                if (etrpType == Const.selective) anyAntiEtrpRqst(peer);
                return; 
            }
        }
        p = new Peer(peer, peer_url, scope, tcp);
        peerTable.put(peer, p);
        System.out.println(">>> Add " + peer + " to peerTable");
        sendDAAdverts(p); // exchange peer information
        if (etrpType == Const.selective) {
            AntiEtrpRqst(peer);
        } else {  // complete
            if (!busy) AntiEtrpRqst(peer);
        }
        return;
    }

    //-------------------------------------------------------------
    // (1) create a peering connection to "url"
    // (2) fork a new thread to read messages from this connection
    // (3) send a DAAdvert to "url" via the peering connection
    //-------------------------------------------------------------
    private synchronized slpTcpHandler createPeerConn(String url) {
        slpTcpHandler tcp = null;
        try {
            String addr = Util.url2dname(url);
            int port = Const.port;
            Socket socket = new Socket(addr, port);
            daTcpHandler h = new daTcpHandler(socket, this);
            tcp = h.getTcp();
            h.start();
            System.out.println(">>> Create peering connection to " + url);
            byte[] buf = composer.DAAdvert(0, Const.normal_flag, "en", 
                         myts, myurl, myscope, myattr, myspi);
            tcp.send(buf, buf.length);
        } catch (ConnectException e) { // keep silent
        } catch (Exception e) {
            System.err.println("createPeerConn:"+e);
        }
        return tcp;
    }

    //----------------------------
    // send DAAdverts to peer "p"
    //----------------------------
    private synchronized void sendDAAdverts(Peer p) {
        int i = 0;
        Iterator values = summaryTable.values().iterator();
        while (values.hasNext()) {
            Summary s = (Summary) values.next();
            String fqdn = s.getFQDN();
            if (!fqdn.equalsIgnoreCase(myfqdn) && 
                 (s.getAcceptTS() != 0 || 
                 !fqdn.equalsIgnoreCase(p.getFQDN()) && 
                 peerTable.containsKey(fqdn)) &&
                Util.shareString(s.getScope(), p.getScope(), ",")) {
                byte[] buf = composer.DAAdvert(0, Const.normal_flag, "en", 
                    s.getBootTS(), s.getURL(), s.getScope(), s.getAttr(), "");
                (p.getTcp()).send(buf, buf.length);
                i++;
            }
        }
        System.out.println(">>> Send " + i + " DAAdverts to " + p.getFQDN());
    }

    public synchronized void AntiEtrpRqst(String peer) {
        if (etrpType == Const.complete) busy = true;
        Peer p = (Peer) peerTable.get(peer);
        Vector adaList = new Vector(10);
        Vector atsList = new Vector(10);
        if (etrpType == Const.complete) {
            p.setPflag();
            Iterator v = summaryTable.values().iterator();
            while (v.hasNext()) {
                Summary s = (Summary) v.next();
                adaList.addElement(s.getFQDN());
                atsList.addElement(new Long(s.getAcceptTS()));
            }
        } else {
            Summary s = (Summary) summaryTable.get(peer);
            adaList.addElement(peer);
            atsList.addElement(new Long(s.getAcceptTS()));
        }
        String id = aidList(adaList, atsList);
        System.out.println(">>> Send AntiEtrpRqst to " + peer + ", ID = " + 
                           id + ", xid = " + xid);
        byte[] buf = composer.AntiEtrpRqst(xid, "en", etrpType, adaList,
                                           atsList); 
        (p.getTcp()).send(buf, buf.length);
        rqstLog.put(new Integer(xid), adaList);
        xid = (xid + 1) & 0xFFFF;
    }

    //------------------------------------------------------------
    // anti-entropy rqst to "any" for new states after "ada":"ats"
    // used by selective anti-entropy for handling failed replicas
    //------------------------------------------------------------
    public synchronized void anyAntiEtrpRqst(String ada) {
        Summary s = (Summary) summaryTable.get(ada);
        long ats = s.getAcceptTS();
        String scope = s.getScope();
        Vector vec = s.getPlist();
        Iterator v = peerTable.values().iterator();
        while (v.hasNext()) {
            Peer p = (Peer) v.next();
            if (Util.shareString(p.getScope(), scope, ",") && 
                !vec.contains(p.getFQDN())) {
                s.addPlist(p.getFQDN());
                Vector adaList = new Vector(1);
                Vector atsList = new Vector(1);
                adaList.addElement(ada);
                atsList.addElement(new Long(ats));
                System.out.println(">>> Send AntiEtrpRqst to " + p.getFQDN() + 
                       ", ID = " + ada + ":" + ats + ", xid = " + xid);
                byte[] buf = composer.AntiEtrpRqst(xid, "en", etrpType, 
                                        adaList, atsList); 
                (p.getTcp()).send(buf, buf.length);
                rqstLog.put(new Integer(xid), adaList);
                xid = (xid + 1) & 0xFFFF;
                return;
            }
        }
    }

    private synchronized String aidList(Vector adaList, Vector atsList) { 
        StringBuffer s = new StringBuffer();
        for (int i=0; i<adaList.size(); i++) {
            if (s.length() > 0) s.append(",");
            s.append((String)adaList.elementAt(i) + ":" +
                     (Long)atsList.elementAt(i));
        }
        return s.toString();
    }

    //-------------------------------------------------
    // fork a new thread and send new states to "peer"
    //-------------------------------------------------
    public synchronized void AntiEtrpRqst_Handler(String peer, int rxid,
                        int rAEtype, Vector adaList, Vector atsList) {
        String id = aidList(adaList, atsList);
//      append("Receive AntiEtrpRqst from " + peer + ", ID = " + id);
        Peer p = (Peer) peerTable.get(peer);
        if (p == null) return;  // not a peer at all
        if (hasNewStates(adaList, atsList)) {
            Vector tmp = database.findNewStates(p.getScope(), adaList,
                                                atsList, rAEtype); 
            (new bulkTrans(p.getTcp(), tmp, peer, rxid, adaList, atsList,
                           this)).start();
        } else {
            AntiEtrpRqst_Handler2(peer, rxid, adaList, atsList);
        }
    }

    private boolean hasNewStates(Vector rdaList, Vector rtsList) {
        Iterator v = summaryTable.values().iterator();
        while (v.hasNext()) {
           Summary s = (Summary) v.next();
           String ada = s.getFQDN();
           long   ats = s.getAcceptTS();
           int index = rdaList.indexOf(ada);
           long   rts   = 0;
           if (index != -1) { // it is a requested subset, find rts
                rts = ((Long) rtsList.get(index)).longValue();
           }
           if (index == -1 && etrpType == Const.complete ||
               index != -1 && ats > rts) {
               return true;
           }
        }
        return false;
    }

    //--------------------------------------------------
    // enable direct forwarding & send SrvAck to "peer"
    //--------------------------------------------------
    public synchronized void AntiEtrpRqst_Handler2(String peer, int rxid,
                             Vector rdaList, Vector rtsList) {
        Peer p = (Peer) peerTable.get(peer);
        if (etrpType == Const.selective) {
            if (rdaList.indexOf(myfqdn) != -1) p.setDirectFwd(true);
        } else { // complete
            p.setDirectFwd(true); 
        }
        byte[] buf = composer.SrvAck(rxid, "en", 0);
        (p.getTcp()).send(buf, buf.length);
        String id = aidList(rdaList, rtsList);
//      append("Send SrvAck to " + peer + ", ID = " + id);
    }

    //--------------------------------------------------
    // one anti-entropy is done, goto next anti-entropy
    //--------------------------------------------------
    public synchronized void SrvAck_Handler(String peer, int rxid) {
        long ms = System.currentTimeMillis() - startTime;
        System.out.println(">>> After " + ms + " ms, receive SrvAck from " + 
                peer + ", xid = " + rxid);
        Vector rdaList = (Vector) rqstLog.remove(new Integer(rxid));
        String rda = (String) rdaList.elementAt(0);
        if (etrpType == Const.selective) {
            if (!peerTable.containsKey(rda)) anyAntiEtrpRqst(rda);
        } else {
            busy = false;
            Iterator v = peerTable.values().iterator();
            while (v.hasNext()) {
                Peer p = (Peer) v.next();
                if (!p.getPflag()) {
                    AntiEtrpRqst(p.getFQDN());
                    break;
                }
            }
        }
    }

    public synchronized void KeepAlive_Handler(String peer) {
        Peer p = (Peer) peerTable.get(peer);
        if (p == null) return;  // not a peer at all
        p.resetAliveTimer();
    }

    public synchronized void setSummary(String ada, long ats, String peer) {
        Summary s = (Summary) summaryTable.get(ada);
        if (s == null) return;  // unknown ADA
        if (ada.equalsIgnoreCase(peer)) { // from accept DA
            s.setAcceptTS(ats);
        } else {
            Summary p = (Summary) summaryTable.get(peer);
            String cs = Util.commonString(s.getScope(), p.getScope(), ",");
            cs = Util.commonString(cs, myscope, ",");
            s.setAcceptTS(ats, cs);
        }
    }

    public void checkPeerConn(String peer, Socket socket) {
        Peer p = (Peer) peerTable.get(peer);
        if (p != null) {
            Socket peerSocket = (p.getTcp()).getSocket();
            if (peerSocket.equals(socket)) {
                System.out.println(">>> Remove " + peer + 
                                   ",  peering connection closed");
                p.closeTcp();
                peerTable.remove(peer);
            } 
        }
    }

    //----------------------------------------------------
    // Forward the message in "buf" with "len" to "scope"
    //----------------------------------------------------
    public synchronized void forward2Peer(byte[] buf, int len, String scope) {
        Iterator v = peerTable.values().iterator();
        while (v.hasNext()) {
            Peer p = (Peer) v.next();
            if (p.isDirectFwd()                              && // can fwd mesgs
                Util.shareString(p.getScope(), scope, ",")) {   // share scopes
                (p.getTcp()).send(buf, len);
            }
        }
    }

//----------------------------------------------------------------------------
// DISPLAY useful information
//----------------------------------------------------------------------------

    private synchronized void displayPeerState() {
        drawLine();
        append("PeerFQDN \t\t\tScopes \tEnableDirectFwd");
        Iterator values = peerTable.values().iterator();
        while (values.hasNext()) {
            Peer p = (Peer) values.next();
            append(p.getFQDN() + "\t" + p.getScope() + "\t" + p.isDirectFwd());
        }
    }

    private void displayMyState() {
        drawLine();
        append("FQDN:\t\t" + myfqdn);
        append("URL in DAAdvert:\t" + myurl);
        append("Port Number:\t\t" + Const.port);
        append("Serving Scopes:\t" + myscope);
        append("Boot Timestamp:\t" + myts);
        append("Attributes:\t\t" + myattr);
        append("Database File:\t\t" + mydbase);
        append("Summary File:\t\t" + mysummary);
        append("Multicast DAAdvert:\t" + mcast);
        append("Mesh-enhanced:\t" + mesh_enhanced);
        append("Anti-Entropy Type:\t" + etrpType);
    }

    private synchronized void displaySummary() {
        drawLine();
        append("Total Number of States: " + database.size());
        drawLine();
        append("DA-FQDN \t\t\tSummaryTS");
        Iterator values = summaryTable.values().iterator();
        while (values.hasNext()) {
            Summary s = (Summary) values.next();
            if (s.getFQDN().equalsIgnoreCase(myfqdn) || s.getAcceptTS() != 0 || 
                peerTable.containsKey(s.getFQDN())) {
                append(s.getFQDN() + "\t" + s.getAcceptTS());
            }
        }
    }

    private synchronized void displayAdvert() {
        drawLine();
        append("PeerURL \t\t\tScopes \tBootTS \tAttributes");
        Iterator values = summaryTable.values().iterator();
        while (values.hasNext()) {
            Summary s = (Summary) values.next();
            if (s.getFQDN().equalsIgnoreCase(myfqdn) || s.getAcceptTS() != 0 || 
                peerTable.containsKey(s.getFQDN())) {
                append(s.getURL() + "\t" + s.getScope() + "\t" + 
                       s.getBootTS() + "\t" + s.getAttr());
            }
        }
    }

    private void drawLine() {
        String s = "-----------------";
        append(s+s+s+s);
    }

//---------------------------------------------------------------------------
// SAVE & LOAD data (database, summary, DAAdvert)
//---------------------------------------------------------------------------

    private synchronized void saveData() {
        try {
            BufferedWriter o = new BufferedWriter(new FileWriter(mydbase));
            database.saveDatabase(o);
            o.close();
            // Save Summary
            o = new BufferedWriter(new FileWriter(mysummary));
            StringBuffer buf = new StringBuffer();
            Iterator values = summaryTable.values().iterator();
            while (values.hasNext()) {
                Summary s = (Summary) values.next();
                if (buf.length() > 0) buf.append("\n");
                String attr = s.getAttr();
                if (attr.equals("")) attr = "none";
                buf.append(s.getFQDN() + " " + s.getURL() + " " + 
                    s.getScope() + " " + 
                    s.getBootTS() + " " + attr + " " + s.getAcceptTS());
            }
            if (buf.length() > 0) buf.append("\n");
            o.write(buf.toString(), 0, buf.length());
            o.close();
        } catch (Exception e) {
            System.err.println("saveData:"+e);
        }
    }

    private synchronized boolean loadData() {
        try {
            File f = new File(mydbase);
            if (!f.exists()) return false;
            f = new File(mysummary);
            if (!f.exists()) return false;
            database.loadDatabase(mydbase);
            // Load Summary
            String line;
            BufferedReader in =
                new BufferedReader(new FileReader(mysummary));
            while ((line = in.readLine()) != null) {
                StringTokenizer st = new StringTokenizer(line);
                String fqdn  = st.nextToken();
                String url   = st.nextToken();
                String scope = st.nextToken();
                int    bts   = Integer.parseInt(st.nextToken());
                String attr  = st.nextToken();
                if (attr.equals("none")) attr = "";
                long   ats   = Long.parseLong(st.nextToken());
                summaryTable.put(fqdn, 
                               new Summary(fqdn, url, scope, bts, attr, ats));
            }
            in.close();
        } catch (Exception e) {
            System.err.println("loadData:"+e.getMessage());
        }
        return true;
    }

//-------------------------------------------------------------------------
// CREATE GUI
//-------------------------------------------------------------------------

    private void createGUI() {
        fileMenu = new JMenu("File");
        exitItem = new JMenuItem("Exit");
        exitItem.addActionListener(this);
        fileMenu.add(exitItem);

        funcMenu = new JMenu("Operation");
        summaryItem = new JMenuItem("Display Summary");
        summaryItem.addActionListener(this);
        funcMenu.add(summaryItem);

        dataItem = new JMenuItem("Display Database");
        dataItem.addActionListener(this);
        funcMenu.add(dataItem);

        advertItem = new JMenuItem("Display DAAdvert");
        advertItem.addActionListener(this);
        funcMenu.add(advertItem);

        funcMenu.addSeparator();
        loadItem = new JMenuItem("Load Data");
        loadItem.addActionListener(this);
        funcMenu.add(loadItem);

        saveItem = new JMenuItem("Save Data");
        saveItem.addActionListener(this);
        funcMenu.add(saveItem);

        setupMenu = new JMenu("Setup");
        peerItem = new JMenuItem("Peering State");
        peerItem.addActionListener(this);
        setupMenu.add(peerItem);

        confItem = new JMenuItem("My Configuration");
        confItem.addActionListener(this);
        setupMenu.add(confItem);

        setupMenu.addSeparator();
        daItem = new JMenuItem("Unicast DAAdvert");
        daItem.addActionListener(this);
        setupMenu.add(daItem);

        helpMenu = new JMenu("Help");
        aboutItem = new JMenuItem("About...");
        aboutItem.addActionListener(this);
        helpMenu.add(aboutItem);

        mbar = new JMenuBar();
        Border raisedbevel, loweredbevel, compound;
        raisedbevel = BorderFactory.createRaisedBevelBorder();
        loweredbevel = BorderFactory.createLoweredBevelBorder();
        compound = BorderFactory.createCompoundBorder(
                                raisedbevel, loweredbevel);
        mbar.setBorder(compound);
        mbar.add(fileMenu);
        mbar.add(funcMenu);
        mbar.add(setupMenu);
        mbar.add(helpMenu);
//      mbar.setHelpMenu(helpMenu);
        setJMenuBar(mbar);

        GridBagLayout gridbag = new GridBagLayout();
        getContentPane().setLayout(gridbag);
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 0.0;
        c.weighty = 0.0;

        urlLabel = new JLabel(" To-URL: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(urlLabel, c);
        getContentPane().add(urlLabel);

        c.weightx = 1.0;
        urlField = new JTextField();
        urlField.setForeground(Color.black);
        urlField.setBackground(Color.lightGray);
        urlField.setBorder(compound);
        c.gridx = 2;
        c.gridy = 0;
        c.gridwidth = 10;
        c.gridheight = 1;
        gridbag.setConstraints(urlField, c);
        getContentPane().add(urlField);

        c.weighty = 1.0;
        textarea = new JTextArea();
        textarea.setEditable(false);
        textarea.setForeground(Color.black);
        textarea.setBackground(Color.lightGray);
        textarea.setBorder(compound);
        JScrollPane textpane = new JScrollPane(textarea);
        c.gridx = 0;
        c.gridy = 1;
        c.gridwidth = 12;
        c.gridheight = 2;
        gridbag.setConstraints(textpane, c);
        getContentPane().add(textpane);
    }
}
