/**
 * Implements the various actions a DA should perform upon receiving 
 * a SLPv2 message: (1) parse message, (2) compose reply
 * 
 * (c) Columbia University, 2001-2004, All Rights Reserved.
 * Author: Weibin Zhao
 */

import java.net.*;
import java.util.*;

public class daAction {

    slpMsgComposer composer;
    da             daf;
    slpTcpHandler  tcp;
    slpUdpHandler  udp;
    boolean        mesh_enhanced;

    daAction(da daf, slpTcpHandler tcp, slpUdpHandler udp) {
        composer = new slpMsgComposer();
        this.daf = daf;
        this.tcp = tcp;
        this.udp = udp;
        mesh_enhanced = daf.getMeshFlag();
    }

    //------------------------------------
    // buf: whole message (header + body)
    // parser: header has been parsed
    //------------------------------------
    public void action(slpMsgParser parser, byte[] buf) {
        byte[] reply, tmp;
        String s, peer, peer_ip, peer_url;
        int id, err;
        id = parser.getFuncID();
        if (tcp != null) {
            peer    = tcp.getPeerName();
            peer_ip = tcp.getPeerIP();
        } else {
            peer    = udp.getPeerName();
            peer_ip = udp.getPeerIP();
        }
        peer_url = Const.DAAdvert_Rqst+"://"+peer_ip;
        int ia[] = { Const.header_len }; // skip header parsing
        parser.LangTag(buf, ia);
        switch (id) {
            case Const.SrvReg:
            case Const.SrvDeReg: 
                parser.MeshFwdExt(buf, peer); // parse extension first  
                if (id == Const.SrvReg) {
                    parser.SrvReg(buf, ia);
                } else {
                    parser.SrvDeReg(buf, ia);
                }
                if (mesh_enhanced && parser.getEcode() == 0 &&
                    parser.getMeshFwdID() == Const.RqstFwd) {
                    daf.forward2Peer(buf, parser.getPacketLen(), 
                                     parser.getScope());
                }
                if (mesh_enhanced && parser.getMeshFwdID() == Const.Fwded) {
                    reply = null;       // no need for SrvAck for efficience
                } else {
                    reply = composer.SrvAck(parser.getXID(), parser.getLtag(), 
                                        parser.getEcode());
                }
                break;
            case Const.SrvRqst: 
                parser.SelectSortExt(buf); // parse extension first
                parser.AttrListExt(buf); // parse extension first
                tmp = parser.SrvRqst(buf, ia);
                if (tmp == null) {
                    daf.append("DAAdvert-Request from " + peer);
                    reply = composer.DAAdvert(parser.getXID(), 
                        Const.normal_flag, "en", daf.getTS(), daf.getURL(), 
                        daf.getScope(), daf.getAttr(), daf.getSPI());
                } else {
                    daf.append("SrvRqst from " + peer);
                    if ((parser.getFlag() & Const.mcast_flag) != 0 &&
                        !Util.shareString(daf.getScope(), 
                                          parser.getScope(), ",")) {
                        reply = null;   // mcast & unsupported scope
                    } else {
                        reply = composer.SrvReply(parser.getXID(),
                                         parser.getLtag(), tmp);
                        if (parser.hasSelectExt()) {
                            reply = composer.SelectExt(reply, 
                                         parser.getTotalMatch());
                        }
                        if (parser.hasAttrListExt()) {
                            Vector matchedEntry = parser.getMatchedEntry();
                            for (int i=0; i<matchedEntry.size(); i++) {
                                Entry e = (Entry) matchedEntry.elementAt(i);
                                reply = composer.AttrListExt(reply, e);
                            }
                        }
                    }
                }
                break;
            case Const.SrvTypeRqst: 
                daf.append("SrvTypeRqst from " + peer);
                parser.SrvTypeRqst(buf, ia);
                if ((parser.getFlag() & Const.mcast_flag) != 0 &&
                    !Util.shareString(daf.getScope(), parser.getScope(), ",")) {
                    reply = null;       // mcast & unsupported scope
                } else {
                    reply = composer.SrvTypeReply(parser.getXID(),
                                parser.getLtag(), parser.getEcode(), 
                                parser.getTypeList());
                }
                break;
            case Const.AttrRqst: 
                daf.append("AttrRqst from " + peer);
                parser.AttrRqst(buf, ia);
                if ((parser.getFlag() & Const.mcast_flag) != 0 &&
                    !Util.shareString(daf.getScope(), parser.getScope(), ",")) {
                    reply = null;       // mcast & unsupported scope
                } else {
                    reply = composer.AttrReply(parser.getXID(),parser.getLtag(),
                                 parser.getEcode(), parser.getAttrList());
                }
                break;
            case Const.DAAdvert:
                if (mesh_enhanced) { // no actions for non-mesh DA
                    parser.DAAdvert(buf, ia);
                    boolean origAdvert=false; // is it original DAAdvert?
                    try { 
                        String advert_ip = Util.url2dname(parser.getURL());
                        advert_ip = // in case a domain name, to ip addr 
                          InetAddress.getByName(advert_ip).getHostAddress();
                        origAdvert = advert_ip.equals(peer_ip); 
                    } catch (Exception e) {
                        System.err.println(e);
                    }
                    if (parser.getURL().equals(daf.getURL())) { // my advert
                        if (!origAdvert) { // fwded my advert
                            daf.anyAntiEtrpRqst(daf.getFQDN());
                        }
                    } else {
                        slpTcpHandler pc = null; // has peering connection?
                        if (tcp != null && origAdvert) pc = tcp;
                        daf.DAAdvertHandler(peer, parser.getURL(), 
                            parser.getScope(), 
                            parser.getAttr(), parser.getDaBootTS(), pc, buf, 
                            origAdvert);  // original advert
                    }
                }
                reply = null;
                break;
            case Const.AntiEtrpRqst:
                parser.AntiEtrpRqst(buf, ia);   // anti-entropy request
                daf.AntiEtrpRqst_Handler(peer, parser.getXID(), 
                                parser.getEtrpType(), parser.getAdaList(), 
                                parser.getAtsList());
                reply = null;
                break;
            case Const.SrvAck: 
                daf.SrvAck_Handler(peer, parser.getXID()); // anti-entropy-end
                reply = null;
                break;
            default:
                System.err.println("Unknown type for SLP");
                reply = null;
                break;
        }
        if (reply != null) {
            if (tcp != null) tcp.send(reply, reply.length);
            if (udp != null) udp.send(reply, peer, udp.getPort());
        }
    }
}
